#include "mcc_generated_files/mcc.h"

#define TIME_OFF -1     //タイマーを使用しない場合
#define TIME_UP 0       //タイムアップ
#define TIM_RX_WAIT 20  //200ms
#define TIM_TX_WAIT 10  //100ms
#define TIM_LED_WAIT 25 //250ms
#define RING_SZ 64
#define SZ_MASK 0x3F
#define HEADER_A 'A'
#define HEADER_P 'P'
#define OFFSET_SZ 3
#define TX_SZ 7

typedef struct{
    uint8_t wp;
    uint8_t rp;
    uint8_t buf[RING_SZ];
}RING_MNG;

typedef struct{
    uint8_t cnt;
    uint8_t data[TX_SZ];
}TX_MNG;

typedef enum{	//ﾓｰﾄﾞ
	COM_RX_WAIT = 0,
	COM_RCVDATA,
	COM_TX,
	COM_MAX
}COM_MODE_NO;

/* 変数定義 */
int8_t      timRxWait = TIME_OFF;
int8_t      timTxWait = TIME_OFF;
int8_t      timLedWait;
RING_MNG    picRcv;
COM_MODE_NO ComMode;
uint8_t     Rcvdata[RING_SZ];
TX_MNG      Txdata;

/* 関数ﾌﾟﾛﾄﾀｲﾌﾟ宣言 */
void eUsartmain(void);
uint8_t CalcSum(uint8_t *buf, uint8_t sz); 
void mainTimer(void); //ﾀｲﾏｰ0割り込みでの処理
void ReadPointerAdd(void);
bool RcvChk(void);
void TxDataSet(uint8_t no);
void mainApp(void);

/* Main application */
void main(void)
{
    SYSTEM_Initialize();
    TMR0_SetInterruptHandler(mainTimer);
    INTERRUPT_GlobalInterruptEnable();
    INTERRUPT_PeripheralInterruptEnable();

    while (1)
    {   
        eUsartmain();
        mainApp();
    }
}
/* タイマ管理関数 */
void mainTimer(void){
 
    if( timRxWait > TIME_UP ){
        --timRxWait; //タイマを更新
    }
    
    if( timTxWait > TIME_UP ){
        --timTxWait; //タイマを更新
    }
    
    if( timLedWait > TIME_UP){
        --timLedWait;
    }
}
/* メイン関数 */
void mainApp(void){
    
    while(EUSART_is_rx_ready())
    {
        picRcv.buf[picRcv.wp] = EUSART_Read(); //データをリード
        IO_RA7_Toggle();
        if( ++picRcv.wp >= sizeof(picRcv.buf)){ //次に保管する位置を更新
            picRcv.wp = 0;
        }
    }
    
    if( timLedWait == TIME_UP){
        timLedWait = TIM_LED_WAIT;
        IO_RA3_Toggle();
    }
}
/* シリアル通信メイン処理 */
void eUsartmain(void){
    int8_t  rxsz;
    uint8_t datsz;
    uint8_t allsz;
    uint8_t rp = picRcv.rp;  
    uint8_t i;
    uint8_t sumchk;
    uint8_t dat[2];
    
    if( timRxWait == TIME_UP){ //受信タイムアウトか
    	timRxWait = TIME_OFF;
	ReadPointerAdd(); 
        ComMode = COM_RX_WAIT;
    }
    
    switch( ComMode ){
        case COM_RX_WAIT:
            rxsz = picRcv.wp - picRcv.rp; //受信データ数の算出
    
            if( rxsz < 0 ){
                rxsz = rxsz + sizeof(picRcv.buf);
            }
            
            if( rxsz == 0 ){
		timRxWait = TIME_OFF;
	    }
            else{
                if( timRxWait == TIME_OFF ){
		    timRxWait = TIM_RX_WAIT; //受信タイムアウトをセット
		}
                
                if( picRcv.buf[ rp ] != HEADER_P  ){ //ヘッダーが一致するか
                    ReadPointerAdd(); //読み込み位置の更新
                }
                else{
                    if( rxsz >= 2 ){ //データ長分だけ獲得しているか
                        for( i = 0; i < 2; i++){//データサイズ算出のため仮おき
                            dat[i] = picRcv.buf[ rp ];
                            if(++rp >= sizeof(picRcv.buf) ) rp = 0;
                        }
                        
                        if( dat[1] > sizeof(picRcv.buf) - OFFSET_SZ ){
                            allsz = sizeof(picRcv.buf) - OFFSET_SZ;
                        }
                        else{
                            datsz = dat[1];
                            allsz = dat[1] + OFFSET_SZ; //header,datalengh,sumを含む
                        }
                        
                        if( rxsz >= allsz){ //確定したサイズ以上か
                            timRxWait = TIME_OFF;
                        
                            for(i=0; i < sizeof(Rcvdata); i++){
                                Rcvdata[i] = 0; //バッファのクリア
                            }
                        
                            for( i=0; i < allsz; i++ ){
                                Rcvdata[i] = picRcv.buf[picRcv.rp]; //一時保管したデータを移す
                                ReadPointerAdd(); //読み込み位置の更新
                            }
                       
                            sumchk = CalcSum(&Rcvdata[2], datsz); //チェックサムの計算
                            if( sumchk == Rcvdata[allsz-1]){ //チェックサムの確認
                                ComMode = COM_RCVDATA; //モードを遷移
                            }
                        }
		    }               
                }
            }
            break;
        case COM_RCVDATA: 
            if( RcvChk() ){
                ComMode = COM_TX;
                timTxWait = TIM_TX_WAIT;
            }
            else{
                ComMode = COM_RX_WAIT;
            }
            break;
        case COM_TX:
            if(EUSART_is_tx_ready()){ //送信ビジーでないか
                if ( Txdata.cnt < sizeof(Txdata.data)){
                    EUSART_Write(Txdata.data[Txdata.cnt]); //データを送信
                    ++Txdata.cnt; //送信データカウントを更新
                }
                else{ //全データの送信が完了
                    timTxWait = TIME_OFF;
                    ComMode = COM_RX_WAIT; //受信待ちに遷移
                }
            }  
                
            if( timTxWait == TIME_UP ){ //送信タイムアウトか
                timTxWait = TIME_OFF;
                ComMode = COM_RX_WAIT; //受信待ちに遷移
            }
            break;        
    }
}
/* 読み込み位置の更新 */
void ReadPointerAdd(void){
    
    if(++picRcv.rp >= sizeof(picRcv.buf) ){
        picRcv.rp = 0;
    }
}
/* チェックサムの計算 */
uint8_t CalcSum(uint8_t *buf, uint8_t sz ){
    uint8_t ret = 0;

    for(uint8_t i=0; i < sz; i++ ){
        ret += *buf;
        buf++; 
    }
    return ret;
}
/* 電文の処理 */
bool RcvChk(void){
    bool ret = false;
    
    if( Rcvdata[2] == 0x30 && Rcvdata[3] == 0x31 && 
        Rcvdata[4] == 0x32 && Rcvdata[5] == 0x33 ){
        IO_RB6_SetHigh();
        TxDataSet(1);
        ret = true;
    }
    else if( Rcvdata[2] == 0x33 && Rcvdata[3] == 0x32 &&
             Rcvdata[4] == 0x31 && Rcvdata[5] == 0x30 ){
        IO_RB6_SetLow();
        TxDataSet(2);
        ret = true;
    }
    return ret;
}
/* 送信データ生成 */
void TxDataSet(uint8_t no){
    
    Txdata.cnt = 0;
    Txdata.data[0] = HEADER_A;
    Txdata.data[1] = 4;
    Txdata.data[2] = 'O';
    Txdata.data[3] = 'K';
    Txdata.data[4] = '-';
    Txdata.data[5] = 0x30 + no;
    Txdata.data[6] = CalcSum(&Txdata.data[2], 4);
}
/**
 End of File
*/